loadmodule "luahandle"
loadmodule "luapsx"
loadmodule "luacd"
loadmodule "lualibs"

load "ips-patcher.lua"

local function buffer_pad(buffer, padding, val)
    local rest, i

    if (not padding) then
        rest = 2048 - andB(buffer:getsize(), 0x7ff)
        if (rest == 2048) then
            return
        end
    else
        rest = padding - math.mod(buffer:getsize(), padding)
    end

    buffer:wseek(0, SEEK_END)
    for i = 1, rest, 1 do
        buffer:writeU8(val or 0)
    end
end

local function process_bs_frame(bs_frame, frameno, framesize, width, height, iso, config)
    local bs_length, bs_magic, bs_qscale, bs_type, bs_pad = bs_frame:readU16(), bs_frame:readU16(), bs_frame:readU16(), bs_frame:readU16(), 0
    if bs_magic ~= 0x3800 then error "Wrong BS file format." end
    bs_frame:seek(0)
    local rgb_picture = nil
    local sect_out = Buffer(true)
    
    print("Processing frame number " .. frameno)

    if not config.stateful then
        config.stateful = {}
        if config.sub then
            load(config.sub.index)
            config.stateful.lastframe = -1
            config.stateful.frame = Buffer()
        end
        if config.sub_karaoke then
            load(config.sub_karaoke.index)
            config.stateful.karaoke_frame = Buffer()
            config.stateful.karaoke_lastframe = 99999999
        end
    end
    
    if frames and frames[frameno] then
        print("Putting normal image")
        if not rgb_picture then rgb_picture = bsdecode(bs_frame, width, height) end
        local sframe = frames[frameno]
        if config.stateful.lastframe ~= sframe then
            print("Opening image " .. sframe)
            config.stateful.frame:destroy()
            config.stateful.frame = Input(config.sub.dir .. "/sub-" .. sframe .. ".rgba")
            config.stateful.lastframe = sframe
        end
        local pos = sub_pos[sframe]
        config.stateful.frame:seek(0)
        print("Blitting normal image")
        blit(rgb_picture, config.stateful.frame, width, height, pos.w, pos.h, pos.x, pos.y, BLIT_ALPHA)
        rgb_picture:seek(0)
        rgb_picture:wseek(0)
    end
    
    if karaoke_frames and karaoke_frames[frameno] then
        print("Putting karaoke image")
        if not rgb_picture then rgb_picture = bsdecode(bs_frame, width, height) end
        local sframe = karaoke_frames[frameno]
        if config.stateful.karaoke_lastframe ~= sframe then
            print("Patching from image " .. config.stateful.karaoke_lastframe .. " to image " .. sframe)
            if config.stateful.karaoke_lastframe > sframe then
                print("Have to reset back from image 1")
                config.stateful.karaoke_lastframe = 1
                config.stateful.karaoke_frame:destroy()
                local b = Buffer(true)
                local i = Input(config.sub_karaoke.dir .. "/sub-1.rgba")
                b:copyfrom(i)
                config.stateful.karaoke_frame = b
                i:destroy()
            end
            for f = config.stateful.karaoke_lastframe + 1, sframe do
                config.stateful.karaoke_frame:wseek(0)
                config.stateful.karaoke_frame:seek(0)
                print("Applying patch " .. f)
                local ips = Input(config.sub_karaoke.dir .. "/sub-" .. f .. ".ips")
                ips_patch(config.stateful.karaoke_frame, ips)
                ips:destroy()
            end
            config.stateful.karaoke_lastframe = sframe
        end
        local pos = karaoke_sub_pos[sframe]
        config.stateful.karaoke_frame:seek(0)
        print("Blitting karaoke image")
        blit(rgb_picture, config.stateful.karaoke_frame, width, height, pos.w, pos.h, pos.x, pos.y, BLIT_ALPHA)
        rgb_picture:seek(0)
        rgb_picture:wseek(0)
    end
    
    if rgb_picture then
        print("Reencoding frame...")
        bs_frame = bsencode(rgb_picture, width, height, 8 * 2016)
        buffer_pad(bs_frame, 8 * 2016, 0xff)
    end
    
    bs_length, bs_magic, bs_qscale, bs_type, bs_pad = bs_frame:readU16(), bs_frame:readU16(), bs_frame:readU16(), bs_frame:readU16(), 0
    bs_frame:seek(0)
    
    for i = 0, 7 do
        sect_out:reset()
        sect_out:writeU16(0x160)
        sect_out:writeU16(0x8001)
        sect_out:writeU16(i)
        sect_out:writeU16(8)
        sect_out:writeU32(frameno)
        sect_out:writeU32(bs_frame:getsize())
        sect_out:writeU16(width)
        sect_out:writeU16(height)
        sect_out:writeU16(bs_length)
        sect_out:writeU16(bs_magic)
        sect_out:writeU16(bs_qscale)
        sect_out:writeU16(bs_type)
        sect_out:writeU32(0)
        sect_out:copyfrom(bs_frame, 2016)
        iso:createsector(sect_out, MODE2_FORM1, config.pos)
        config.pos = config.pos + 1
    end
    
    if rgb_picture then
        rgb_picture:destroy()
        bs_frame:destroy()
    end
end

local function process_sector(str, bs_frame, iso, config)
    local r = {}
    local frame_done = true
    r.Status = str:readU16()
    r.Type = str:readU16()
    r.SectorOffset = str:readU16()
    r.SectorSize = str:readU16()
    r.FrameNo = str:readU32()
    r.FrameSize = str:readU32()
    r.FrameWidth = str:readU16()
    r.FrameHeight = str:readU16()
    r.bs_length = str:readU16()
    r.bs_magic = str:readU16()
    r.bs_q_scale = str:readU16()
    r.bs_type = str:readU16()
    r.bs_pad = str:readU32()
    
    if r.Status ~= 0x160 then error("Seems we got a wrong header: " .. r.Status) end
    if r.Type == 1 then
        if not frame_done then error "Interlaced audio in video." end
        str:seek(-32, SEEK_CUR)
        iso:createsector(str, MODE2_FORM1, config.pos)
        config.pos = config.pos + 1
    elseif r.Type == 0x8001 then
        if r.SectorSize ~= 8 then error "video frame less than 8 sectors..." end
        
        frame_done = r.SectorOffset == r.SectorSize - 1
        bs_frame:copyfrom(str, 2016)
        if frame_done then return true, r.FrameNo, r.FrameSize, r.FrameWidth, r.FrameHeight end
    else
        error("Unknown sector type: " .. r.Type)
    end
    
    return false
end

local function process_lom_str(str, iso, config)
    local num_sect = str:getsize() / 2048
    local bs_frame = Buffer(true)
    local frame_done = true
    local frame_no
    local frame_size
    local frame_width
    local frame_height
    
    for i = 0, num_sect - 1 do
        str:seek(i * 2048)
        frame_done, frame_no, frame_size, frame_width, frame_height = process_sector(str, bs_frame, iso, config)
        if frame_done then
            process_bs_frame(bs_frame, frame_no, frame_size, frame_width, frame_height, iso, config)
            bs_frame:reset()
        end
    end
end

function insert_str(file_in, filename, path, cdutil, iso, direntry, dirtree, config)
    config.pos = iso:getdispsect()
    local t = Buffer()

    iso:createfile(dirtree, filename, t, direntry).size = file_in:getsize()
    t:destroy()
    process_lom_str(file_in, iso, config)
    config.stateful = nil
    config.pos = nil
    sub_pos = nil
    frames = nil
    karaoke_sub_pos = nil
    karaoke_frames = nil
    
    return file_in:getsize()
end

--------------------------------------------------------------------------------

loadmodule "luacd"
i = cdabstract "lom.bin"
cdutil = cdutils(i)
direntry = cdutil:findpath "/MOVIE/POST.STR;1"
str = cdutil:cdfile(direntry)
config = {
    sub = { index = "subs-post/index.lua", dir = "subs-post" },
    sub_karaoke = { index = "karaoke/index.lua", dir = "karaoke" },
}
iso_file_out = Output "blah.bin"
iso_out = isobuilder(iso_file_out)
iso_out:foreword(cdutil)
pvd = createpvd(cdutil)
root = iso_out:setbasics(pvd)
root:fromdir(cdutil:findpath "/")
insert_str(str, "POST.STR", "/", cdutil, iso_out, direntry, root, config)
iso_out:close()
